@@ -1,4 +1,7 @@
-from gi.repository import Gio
+from datetime import datetime
+from dateutil import tz
+from gettext import gettext as _
+from gi.repository import Gio, GLib
 from pykeepass.kdbx_parsing.kdbx import KDBX
 from pykeepass import PyKeePass
 import hashlib
@@ -317,6 +320,7 @@ class DatabaseManager:
     def add_group_to_database(self, name, icon, notes, parent_group):
         group = self.db.add_group(parent_group, name, icon=icon, notes=notes)
         self.changes = True
+        self.set_element_mtime(parent_group)
 
         return group
 
@@ -324,6 +328,8 @@ class DatabaseManager:
     def delete_group_from_database(self, group):
         self.db.delete_group(group)
         self.changes = True
+        if group.parentgroup is not None:
+            self.set_element_mtime(group.parentgroup)
 
     # Add new entry to database
     def add_entry_to_database(
@@ -333,6 +339,7 @@ class DatabaseManager:
             destination_group, name, username, password, url=url, notes=notes,
             expiry_time=None, tags=None, icon=icon, force_creation=self.check_entry_in_group_exists("", destination_group))
         self.changes = True
+        self.set_element_mtime(destination_group)
 
         return entry
 
@@ -340,26 +347,52 @@ class DatabaseManager:
     def delete_entry_from_database(self, entry):
         self.db.delete_entry(entry)
         self.changes = True
+        if entry.parentgroup is not None:
+            self.set_element_mtime(entry.parentgroup)
+
+    # Duplicate an entry
+    def duplicate_entry(self, entry):
+        title = entry.title
+        if title is None:
+            title = ""
+
+        username = entry.username
+        if username is None:
+            username = ""
+
+        password = entry.password
+        if password is None:
+            password = ""
+
+        # NOTE: With clone is meant a duplicated object, not the process of cloning/duplication; "the" clone
+        clone_entry = self.db.add_entry(entry.parentgroup, title + " - " + _("Clone"), username, password, url=entry.url, notes=entry.notes, expiry_time=entry.expiry_time, tags=entry.tags, icon=entry.icon, force_creation=True)
+
+        # Add custom properties
+        for key in entry.custom_properties:
+            value = entry.custom_properties[key]
+            if value is None:
+                value = ""
+            clone_entry.set_custom_property(key, value)
+
+        self.changes = True
+        if entry.parentgroup is not None:
+            self.set_element_mtime(entry.parentgroup)
 
     # Write all changes to database
     def save_database(self):
         if self.save_running is False and self.changes is True:
             self.save_running = True
 
-            kdbx = KDBX.build(self.db.kdbx, password=self.db.password, keyfile=self.db.keyfile)
-
             try:
-                output_stream = self.database_file_descriptor.replace(None, False, Gio.FileCreateFlags.REPLACE_DESTINATION | Gio.FileCreateFlags.PRIVATE, None)
-                output_stream.write_all(kdbx)
-                output_stream.close()
-                self.logging_manager.log_debug("Saved database")
+                self.db.save()
+                self.logging_manager.debug("Saved database")
                 self.changes = False
             except Exception:
-                self.logging_manager.log_error("Error occured while saving database")
+                self.logging_manager.error("Error occured while saving database")
 
             # Workaround
             # Fix created and proposed: https://github.com/pschmitt/pykeepass/pull/102
-            self.db.kdbx = KDBX.parse_file(self.db.filename, password=self.db.password, keyfile=self.db.keyfile)
+            self.db.kdbx = KDBX.parse_file(self.db.filename, password=self.db.password, keyfile=self.db.keyfile, transformed_key=None)
 
             self.save_running = False
 
@@ -381,57 +414,79 @@ class DatabaseManager:
         entry = self.db.find_entries(uuid=uuid, first=True)
         entry.title = name
         self.changes = True
+        self.set_element_mtime(entry)
 
     def set_entry_username(self, uuid, username):
         entry = self.db.find_entries(uuid=uuid, first=True)
         entry.username = username
         self.changes = True
+        self.set_element_mtime(entry)
 
     def set_entry_password(self, uuid, password):
         entry = self.db.find_entries(uuid=uuid, first=True)
         entry.password = password
         self.changes = True
+        self.set_element_mtime(entry)
 
     def set_entry_url(self, uuid, url):
         entry = self.db.find_entries(uuid=uuid, first=True)
         entry.url = url
         self.changes = True
+        self.set_element_mtime(entry)
 
     def set_entry_notes(self, uuid, notes):
         entry = self.db.find_entries(uuid=uuid, first=True)
         entry.notes = notes
         self.changes = True
+        self.set_element_mtime(entry)
 
     def set_entry_icon(self, uuid, icon):
         entry = self.db.find_entries(uuid=uuid, first=True)
         entry.icon = icon
         self.changes = True
+        self.set_element_mtime(entry)
 
     def set_entry_expiry_date(self, uuid, date):
         entry = self.db.find_entries(uuid=uuid, first=True)
         entry.expiry_time = date
         entry.expires
         self.changes = True
+        self.set_element_mtime(entry)
 
     def set_entry_color(self, uuid, color):
         entry = self.db.find_entries(uuid=uuid, first=True)
         entry.set_custom_property("color_prop_LcljUMJZ9X", color)
         self.changes = True
+        self.set_element_mtime(entry)
 
     def set_entry_attribute(self, uuid, key, value):
         entry = self.db.find_entries(uuid=uuid, first=True)
         entry.set_custom_property(key, value)
         self.changes = True
+        self.set_element_mtime(entry)
 
     def delete_entry_attribute(self, uuid, key):
         entry = self.db.find_entries(uuid=uuid, first=True)
         entry.delete_custom_property(key)
         self.changes = True
+        self.set_element_mtime(entry)
 
     # Move an entry to another group
     def move_entry(self, uuid, destination_group_object):
         entry = self.db.find_entries(uuid=uuid, first=True)
         self.db.move_entry(entry, destination_group_object)
+        if entry.parentgroup is not None:
+            self.set_element_mtime(entry.parentgroup)
+        self.set_element_mtime(destination_group_object)
+
+    def set_element_ctime(self, element):
+        element.ctime = datetime.utcnow()
+
+    def set_element_atime(self, element):
+        element.atime = datetime.utcnow()
+
+    def set_element_mtime(self, element):
+        element.mtime = datetime.utcnow()
 
     #
     # Group Modifications
@@ -441,19 +496,25 @@ class DatabaseManager:
         group = self.db.find_groups(uuid=uuid, first=True)
         group.name = name
         self.changes = True
+        self.set_element_mtime(group)
 
     def set_group_notes(self, uuid, notes):
         group = self.db.find_groups(uuid=uuid, first=True)
         group.notes = notes
+        self.set_element_mtime(group)
 
     def set_group_icon(self, uuid, icon):
         group = self.db.find_groups(uuid=uuid, first=True)
         group.icon = icon
+        self.set_element_mtime(group)
 
     # Move an group
     def move_group(self, uuid, destination_group_object):
         group = self.db.find_groups(uuid=uuid, first=True)
         self.db.move_group(group, destination_group_object)
+        if group.parentgroup is not None:
+            self.set_element_mtime(group.parentgroup)
+        self.set_element_mtime(destination_group_object)
 
     #
     # Read Database
@@ -495,48 +556,30 @@ class DatabaseManager:
         else:
             return True
 
-    # Search for an entry or a group by (part of) name, username, url and notes, returns list of uuid's, search fulltext optionally
-    def global_search(self, string, fulltext):
-        lstring = str.lower(string)
+    # Search for an entry or a group
+    def search(self, string, fulltext, global_search=True, path=None):
         uuid_list = []
-        for group in self.db.groups:
-            if lstring in str.lower(self.get_group_name_from_group_object(group)):
+
+        if fulltext is False:
+            for group in self.db.find_groups(name="(?i)" + string.replace(" ", ".*"), recursive=global_search, path=path, regex=True):
                 if group.is_root_group is False:
                     uuid_list.append(group.uuid)
-            if fulltext is True:
-                if lstring in str.lower(self.get_group_notes_from_group_object(group)):
-                    if group.is_root_group is False:
+        else:
+            for group in self.db.groups:
+                if group.is_root_group is False and group.uuid not in uuid_list:
+                    if string.lower() in self.get_group_notes_from_group_object(group):
                         uuid_list.append(group.uuid)
 
-        for entry in self.db.entries:
-            if lstring in str.lower(self.get_entry_name_from_entry_object(entry)):
+        if fulltext is False:
+            for entry in self.db.find_entries(title="(?i)" + string.replace(" ", ".*"), recursive=global_search, path=path, regex=True):
                 uuid_list.append(entry.uuid)
-            if fulltext is True:
-                if lstring in str.lower(self.get_entry_username_from_entry_object(entry)) or string in str.lower(self.get_entry_url_from_entry_object(entry)) or string in str.lower(self.get_entry_notes_from_entry_object(entry)):
-                   uuid_list.append(entry.uuid) 
-
-        return uuid_list
-
-    # Search one group for a string, search fulltext optionally, returns list of uuid's of groups and entries
-    def local_search(self, group, string, fulltext):
-        lstring = str.lower(string)
-        uuid_list = []
-
-        for group_in in group.subgroups:
-            if lstring in str.lower(self.get_group_name_from_group_object(group_in)):
-                if group_in.is_root_group is False:
-                    uuid_list.append(group_in.uuid)
-            if fulltext is True:
-                if lstring in str.lower(self.get_group_notes_from_group_object(group_in)):
-                    if group_in.is_root_group is False:
-                        uuid_list.append(group_in.uuid)
-
-        for entry in group.entries:
-            if lstring in str.lower(self.get_entry_name_from_entry_object(entry)):
-                uuid_list.append(entry.uuid)
-            if fulltext is True:
-                if lstring in str.lower(self.get_entry_username_from_entry_object(entry)) or string in str.lower(self.get_entry_url_from_entry_object(entry)) or string in str.lower(self.get_entry_notes_from_entry_object(entry)):
-                   uuid_list.append(entry.uuid)
+        else:
+            for entry in self.db.entries:
+                if entry.uuid not in uuid_list:
+                    if string.lower() in self.get_entry_username_from_entry_object(entry):
+                        uuid_list.append(entry.uuid)
+                    elif string.lower() in self.get_entry_notes_from_entry_object(entry):
+                        uuid_list.append(entry.uuid)
 
         return uuid_list
 
@@ -547,6 +590,54 @@ class DatabaseManager:
         else:
             return True
 
+    def check_is_group_object(self, group):
+        return hasattr(group, "name")
+
+    #
+    # Properties
+    #
+
+    def get_element_creation_date(self, element):
+        if element.ctime is not None:
+            local_timestamp = element.ctime.astimezone(tz.tzlocal())
+            timestamp = GLib.DateTime.new_local(
+                int(datetime.strftime(local_timestamp, "%Y")),
+                int(datetime.strftime(local_timestamp, "%m")),
+                int(datetime.strftime(local_timestamp, "%d")),
+                int(datetime.strftime(local_timestamp, "%H")),
+                int(datetime.strftime(local_timestamp, "%M")),
+                float(datetime.strftime(local_timestamp, "%S")))
+            return timestamp.format("%c")
+        else:
+            return "-"
+
+    def get_element_acessed_date(self, element):
+        if element.atime is not None:
+            local_timestamp = element.atime.astimezone(tz.tzlocal())
+            timestamp = GLib.DateTime.new_local(
+                int(datetime.strftime(local_timestamp, "%Y")),
+                int(datetime.strftime(local_timestamp, "%m")),
+                int(datetime.strftime(local_timestamp, "%d")),
+                int(datetime.strftime(local_timestamp, "%H")),
+                int(datetime.strftime(local_timestamp, "%M")),
+                float(datetime.strftime(local_timestamp, "%S")))
+            return timestamp.format("%c")
+        else:
+            return "-"
+
+    def get_element_modified_date(self, element):
+        if element.mtime is not None:
+            local_timestamp = element.mtime.astimezone(tz.tzlocal())
+            timestamp = GLib.DateTime.new_local(
+                int(datetime.strftime(local_timestamp, "%Y")),
+                int(datetime.strftime(local_timestamp, "%m")),
+                int(datetime.strftime(local_timestamp, "%d")),
+                int(datetime.strftime(local_timestamp, "%H")),
+                int(datetime.strftime(local_timestamp, "%M")),
+                float(datetime.strftime(local_timestamp, "%S")))
+            return timestamp.format("%c")
+        else:
+            return "-"
     #
     # Database creation methods
     #
@@ -600,4 +691,4 @@ class DatabaseManager:
 
     def get_database_derivation(self):
         return self.db.version
-        
+
