@@ -34,7 +34,7 @@ class MainWindow(Gtk.ApplicationWindow):
 
         self.assemble_window()
 
-        if self.logging_manager.development_mode is True:
+        if Gio.Application.get_default().development_mode is True:
             passwordsafe.config_manager.set_development_backup_mode(True)
 
     def assemble_window(self):
@@ -70,7 +70,7 @@ class MainWindow(Gtk.ApplicationWindow):
 
         self.set_titlebar(self.headerbar)
 
-        if self.logging_manager.development_mode is True:
+        if Gio.Application.get_default().development_mode is True:
             context = self.get_style_context()
             context.add_class("devel")
 
@@ -182,11 +182,11 @@ class MainWindow(Gtk.ApplicationWindow):
             for g_file in self.get_application().file_list:
                 self.start_database_opening_routine(g_file.get_basename(), g_file.get_path())
         elif passwordsafe.config_manager.get_first_start_screen() is True and filepath is not "" and filepath is not None and os.path.exists(filepath) is True:
-            self.logging_manager.log_debug("Found last opened database (" + filepath + ")")
+            self.logging_manager.debug("Found last opened database (" + filepath + ")")
             tab_title = ntpath.basename(filepath)
             self.start_database_opening_routine(tab_title, filepath)
         else:
-            self.logging_manager.log_debug(
+            self.logging_manager.debug(
                 "No / Not valid last opened database found.")
             self.assemble_first_start_screen()
 
@@ -294,7 +294,7 @@ class MainWindow(Gtk.ApplicationWindow):
         response = filechooser_opening_dialog.run()
 
         if response == Gtk.ResponseType.ACCEPT:
-            self.logging_manager.log_debug(
+            self.logging_manager.debug(
                 "File selected: " + filechooser_opening_dialog.get_filename())
 
             database_already_opened = False
@@ -312,7 +312,7 @@ class MainWindow(Gtk.ApplicationWindow):
                 self.start_database_opening_routine(
                     tab_title, filechooser_opening_dialog.get_filename())
         elif response == Gtk.ResponseType.CANCEL:
-            self.logging_manager.log_debug("File selection canceled")
+            self.logging_manager.debug("File selection canceled")
 
     def start_database_opening_routine(self, tab_title, filepath):
         builder = Gtk.Builder()
@@ -391,7 +391,7 @@ class MainWindow(Gtk.ApplicationWindow):
         if self.container == NotImplemented:
             self.create_container()
 
-        page_instance = ContainerPage(headerbar, self.logging_manager.development_mode)
+        page_instance = ContainerPage(headerbar, Gio.Application.get_default().development_mode)
 
         tab_hbox = Gtk.HBox.new(False, 0)
         tab_label = Gtk.Label.new(title)
@@ -528,7 +528,7 @@ class MainWindow(Gtk.ApplicationWindow):
     def on_application_quit(self, window, event):
         unsaved_databases_list = []
         for db in self.opened_databases:
-            if db.database_manager.changes is True:
+            if db.database_manager.changes is True and db.database_manager.save_running is False:
                 if passwordsafe.config_manager.get_save_automatically() is True:
                     db.stop_save_loop()
                     save_thread = threading.Thread(target=db.database_manager.save_database)
@@ -599,9 +599,21 @@ class MainWindow(Gtk.ApplicationWindow):
 
     # Entry/Group Row Popover Actions
     def add_row_popover_actions(self):
-        delete_entry_action = Gio.SimpleAction.new("entry.delete", None)
-        delete_entry_action.connect("activate", self.execute_gio_action, "on_entry_delete_menu_button_clicked")
-        self.application.add_action(delete_entry_action)
+        delete_element_action = Gio.SimpleAction.new("element.delete", None)
+        delete_element_action.connect("activate", self.execute_gio_action, "on_element_delete_menu_button_clicked")
+        self.application.add_action(delete_element_action)
+
+        duplicate_entry_action = Gio.SimpleAction.new("entry.duplicate", None)
+        duplicate_entry_action.connect("activate", self.execute_gio_action, "on_entry_duplicate_menu_button_clicked")
+        self.application.add_action(duplicate_entry_action)
+
+        references_entry_action = Gio.SimpleAction.new("entry.references", None)
+        references_entry_action.connect("activate", self.execute_gio_action, "on_entry_references_menu_button_clicked")
+        self.application.add_action(references_entry_action)
+
+        properties_element_action = Gio.SimpleAction.new("element.properties", None)
+        properties_element_action.connect("activate", self.execute_gio_action, "on_element_properties_menu_button_clicked")
+        self.application.add_action(properties_element_action)
 
         edit_group_action = Gio.SimpleAction.new("group.edit", None)
         edit_group_action.connect("activate", self.execute_gio_action, "on_group_edit_menu_button_clicked")
@@ -613,6 +625,14 @@ class MainWindow(Gtk.ApplicationWindow):
 
     # MenuButton Popover Actions
     def add_database_menubutton_popover_actions(self):
+        db_add_entry_action = Gio.SimpleAction.new("db.add_entry", None)
+        db_add_entry_action.connect("activate", self.execute_gio_action, "on_database_add_entry_clicked")
+        self.application.add_action(db_add_entry_action)
+
+        db_add_group_action = Gio.SimpleAction.new("db.add_group", None)
+        db_add_group_action.connect("activate", self.execute_gio_action, "on_database_add_group_clicked")
+        self.application.add_action(db_add_group_action)
+
         db_settings_action = Gio.SimpleAction.new("db.settings", None)
         db_settings_action.connect("activate", self.execute_gio_action, "on_database_settings_entry_clicked")
         self.application.add_action(db_settings_action)
@@ -645,12 +665,22 @@ class MainWindow(Gtk.ApplicationWindow):
         if action_db is NotImplemented:
             return
 
-        if name == "on_entry_delete_menu_button_clicked":
-            action_db.on_entry_delete_menu_button_clicked(action, param)
+        if name == "on_element_delete_menu_button_clicked":
+            action_db.on_element_delete_menu_button_clicked(action, param)
+        elif name == "on_entry_duplicate_menu_button_clicked":
+            action_db.on_entry_duplicate_menu_button_clicked(action, param)
+        elif name == "on_entry_references_menu_button_clicked":
+            action_db.show_references_dialog(action, param)
+        elif name == "on_element_properties_menu_button_clicked":
+            action_db.show_properties_dialog(action, param)
         elif name == "on_group_edit_menu_button_clicked":
             action_db.on_group_edit_menu_button_clicked(action, param)
         elif name == "on_group_delete_menu_button_clicked":
             action_db.on_group_delete_menu_button_clicked(action, param)
+        elif name == "on_database_add_entry_clicked":
+            action_db.on_add_entry_button_clicked(None)
+        elif name == "on_database_add_group_clicked":
+            action_db.on_add_group_button_clicked(None)
         elif name == "on_database_settings_entry_clicked":
             action_db.on_database_settings_entry_clicked(action, param)
         elif name == "on_sort_menu_button_entry_clicked":
