@@ -1,20 +1,24 @@
 from gettext import gettext as _
 from gi.repository import Gio, Gdk, Gtk, GLib, Handy
+from passwordsafe.custom_keypress_handler import CustomKeypressHandler
 from passwordsafe.database_settings_dialog import DatabaseSettingsDialog
 from passwordsafe.entry_page import EntryPage
 from passwordsafe.entry_row import EntryRow
 from passwordsafe.group_page import GroupPage
 from passwordsafe.group_row import GroupRow
 from passwordsafe.pathbar import Pathbar
+from passwordsafe.properties_dialog import PropertiesDialog
+from passwordsafe.references_dialog import ReferencesDialog
 from passwordsafe.responsive_ui import ResponsiveUI
 from passwordsafe.scrolled_page import ScrolledPage
 from passwordsafe.selection_ui import SelectionUI
 from passwordsafe.search import Search
 from threading import Timer
-import datetime
+import codecs
 import ntpath
 import os
 import passwordsafe.config_manager
+import re
 import time
 import threading
 
@@ -29,6 +33,7 @@ class UnlockedDatabase:
     search = NotImplemented
     entry_page = NotImplemented
     group_page = NotImplemented
+    custom_keypress_handler = NotImplemented
 
     # Widgets
     parent_widget = NotImplemented
@@ -43,6 +48,9 @@ class UnlockedDatabase:
     overlay = NotImplemented
     search_overlay = NotImplemented
     database_settings_dialog = NotImplemented
+    references_dialog = NotImplemented
+    notes_dialog = NotImplemented
+    properties_dialog = NotImplemented
 
     # Objects
     builder = NotImplemented
@@ -58,13 +66,6 @@ class UnlockedDatabase:
     dbus_subscription_id = NotImplemented
     listbox_insert_thread = NotImplemented
 
-    entry_marked_for_delete = NotImplemented
-    group_marked_for_delete = NotImplemented
-    group_marked_for_edit = NotImplemented
-
-    entries_selected = []
-    groups_selected = []
-
     def __init__(self, window, widget, dbm, unlock_database):
         # Instances
         self.window = window
@@ -76,6 +77,7 @@ class UnlockedDatabase:
         self.search = Search(self)
         self.entry_page = EntryPage(self)
         self.group_page = GroupPage(self)
+        self.custom_keypress_handler = CustomKeypressHandler(self)
 
         # Declare database as opened
         self.window.opened_databases.append(self)
@@ -83,7 +85,7 @@ class UnlockedDatabase:
         # Browser Mode
         self.assemble_listbox()
         self.start_save_loop()
-        self.register_special_keys()
+        self.custom_keypress_handler.register_custom_keys()
         self.register_dbus_signal()
 
         # Responsive UI
@@ -159,12 +161,6 @@ class UnlockedDatabase:
         back_button_mobile = self.builder.get_object("back_button_mobile")
         back_button_mobile.connect("clicked", self.on_back_button_mobile_clicked)
 
-        add_entry_button = self.builder.get_object("add_entry_button")
-        add_entry_button.connect("clicked", self.on_add_entry_button_clicked)
-
-        add_group_button = self.builder.get_object("add_group_button")
-        add_group_button.connect("clicked", self.on_add_group_button_clicked)
-
         # Search UI
         self.search.initialize()
 
@@ -178,14 +174,13 @@ class UnlockedDatabase:
 
     # Group and entry browser headerbar
     def set_browser_headerbar(self):
-        mod_box = self.builder.get_object("mod_box")
+        self.builder.get_object("linkedbox_right").show_all()
 
-        for child in mod_box.get_children():
-            mod_box.remove(child)
+        filename_label = self.builder.get_object("filename_label")
+        filename_label.set_text(ntpath.basename(self.database_manager.database_path))
 
-        mod_box.add(self.builder.get_object("browser_buttons_box"))
-        mod_box.show_all()
-        self.builder.get_object("linked_box1").show_all()
+        secondary_menupopover_button = self.builder.get_object("secondary_menupopover_button")
+        secondary_menupopover_button.hide()
 
         self.responsive_ui.headerbar_back_button()
         self.responsive_ui.headerbar_selection_button()
@@ -200,60 +195,6 @@ class UnlockedDatabase:
         key, mod = Gtk.accelerator_parse(accelerator)
         widget.add_accelerator(signal, accelerators, key, mod, Gtk.AccelFlags.VISIBLE)
 
-    #
-    # Special Keys (e.g. type-to-search)
-    #
-
-    def register_special_keys(self):
-        self.window.connect("key-release-event", self.on_special_key_pressed)
-
-    def on_special_key_pressed(self, window, eventkey):
-        group_uuid = self.database_manager.get_group_uuid_from_group_object(self.current_group)
-
-        if self.window.container.page_num(self.parent_widget) == self.window.container.get_current_page():
-            scrolled_page = self.stack.get_child_by_name(group_uuid)
-            if self.database_locked is False and self.selection_ui.selection_mode_active is False and self.database_manager.check_is_group(self.database_manager.get_group_uuid_from_group_object(self.current_group)) and scrolled_page.edit_page is False:
-                if self.stack.get_visible_child() is not self.stack.get_child_by_name("search"):
-                    if eventkey.string.isalpha() or eventkey.string.isnumeric():
-                        self.search.set_search_headerbar(self.builder.get_object("search_button"))
-                        self.builder.get_object("headerbar_search_entry").set_text(eventkey.string)
-                        Gtk.Entry.do_move_cursor(self.builder.get_object("headerbar_search_entry"), Gtk.MovementStep.BUFFER_ENDS, 1, False)
-                    elif eventkey.keyval == Gdk.KEY_BackSpace:
-                        uuid = self.stack.get_visible_child_name()
-                        if self.database_manager.check_is_root_group(self.current_group) is False:
-                            if self.database_manager.check_is_root_group(self.database_manager.get_group_parent_group_from_uuid(uuid)) is True:
-                                self.pathbar.on_home_button_clicked(self.pathbar.home_button)
-                            else:
-                                for button in self.pathbar:
-                                    if button.get_name() == "PathbarButtonDynamic" and type(button) is passwordsafe.pathbar_button.PathbarButton:
-                                        if button.uuid == self.database_manager.get_group_uuid_from_group_object(self.database_manager.get_group_parent_group_from_uuid(uuid)):
-                                            self.pathbar.on_pathbar_button_clicked(button)
-            elif self.database_locked is False and self.selection_ui.selection_mode_active is False and self.stack.get_visible_child() is not self.stack.get_child_by_name("search"):
-                if eventkey.keyval == Gdk.KEY_Escape:
-                    uuid = self.stack.get_visible_child_name()
-                    if self.database_manager.check_is_group(uuid):
-                        scrolled_page = self.stack.get_child_by_name(uuid)
-                        if self.database_manager.check_is_root_group(self.database_manager.get_group_parent_group_from_uuid(uuid)) is True:
-                            self.pathbar.on_home_button_clicked(self.pathbar.home_button)
-                        else:
-                            if scrolled_page.edit_page is True:
-                                for button in self.pathbar:
-                                    if button.get_name() == "PathbarButtonDynamic" and type(button) is passwordsafe.pathbar_button.PathbarButton:
-                                        if self.database_manager.get_group_object_from_uuid(uuid) is None:
-                                            return
-                                        if button.uuid == self.database_manager.get_group_uuid_from_group_object(self.database_manager.get_group_parent_group_from_uuid(uuid)):
-                                            self.pathbar.on_pathbar_button_clicked(button)
-                    else:
-                        if self.database_manager.check_is_root_group(self.database_manager.get_entry_parent_group_from_uuid(uuid)) is True:
-                            self.pathbar.on_home_button_clicked(self.pathbar.home_button)
-                        else:
-                            for button in self.pathbar:
-                                if button.get_name() == "PathbarButtonDynamic" and type(button) is passwordsafe.pathbar_button.PathbarButton:
-                                    if self.database_manager.get_entry_object_from_uuid(uuid) is None:
-                                        return
-                                    if button.uuid == self.database_manager.get_group_uuid_from_group_object(self.database_manager.get_entry_parent_group_from_uuid(uuid)):
-                                        self.pathbar.on_pathbar_button_clicked(button)
-
     #
     # Group and Entry Management
     #
@@ -276,7 +217,6 @@ class UnlockedDatabase:
             scrolled_window = ScrolledPage(True)
 
             viewport = Gtk.Viewport()
-            viewport.set_name("BGPlatform")
             scrolled_window.properties_list_box = builder.get_object("properties_list_box")
 
             # Responsive Container
@@ -300,6 +240,7 @@ class UnlockedDatabase:
             self.group_page.set_group_edit_page_headerbar()
         # If the stack page with current group's uuid isn't existing - we need to create it (first time opening of group/entry)
         elif self.stack.get_child_by_name(self.database_manager.get_group_uuid_from_group_object(self.current_group)) is None and self.stack.get_child_by_name(self.database_manager.get_entry_uuid_from_entry_object(self.current_group)) is None and edit_group is False:
+            self.database_manager.set_element_atime(self.current_group)
             # Create not existing stack page for group
             if self.database_manager.check_is_group(self.database_manager.get_group_uuid_from_group_object(self.current_group)) is True:
                 builder = Gtk.Builder()
@@ -309,7 +250,6 @@ class UnlockedDatabase:
 
                 scrolled_window = ScrolledPage(False)
                 viewport = Gtk.Viewport()
-                viewport.set_name("BGPlatform")
                 overlay = Gtk.Overlay()
 
                 # Responsive Container
@@ -342,7 +282,6 @@ class UnlockedDatabase:
                 scrolled_window = ScrolledPage(True)
 
                 viewport = Gtk.Viewport()
-                viewport.set_name("BGPlatform")
                 scrolled_window.properties_list_box = builder.get_object("properties_list_box")
 
                 # Responsive Container
@@ -363,6 +302,7 @@ class UnlockedDatabase:
                     self.entry_page.insert_entry_properties_into_listbox(scrolled_window.properties_list_box, False)
         # Stack page with current group's uuid already exists, we only need to switch stack page
         else:
+            self.database_manager.set_element_atime(self.current_group)
             # For group
             if self.database_manager.check_is_group(self.database_manager.get_group_uuid_from_group_object(self.current_group)) is True:
                 self.stack.set_visible_child_name(self.database_manager.get_group_uuid_from_group_object(self.current_group))
@@ -396,7 +336,7 @@ class UnlockedDatabase:
 
             if stack_page is not None:
                 stack_page.destroy()
-                
+
             self.scheduled_page_destroy.remove(page_uuid)
             self.show_page_of_new_directory(False, False)
 
@@ -518,11 +458,8 @@ class UnlockedDatabase:
             return
 
         if list_box_row.get_type() == "EntryRow" and self.selection_ui.selection_mode_active is True:
-            if list_box_row.selection_checkbox.get_active():
-                list_box_row.selection_checkbox.set_active(False)
-            else:
-                list_box_row.selection_checkbox.set_active(True)
-        elif list_box_row.get_type() == "EntryRow" and self.selection_ui.selection_mode_active is not True:
+            self.selection_ui.row_selection_toggled(list_box_row)
+        elif list_box_row.get_type() == "EntryRow" and self.selection_ui.selection_mode_active is False:
             self.set_current_group(self.database_manager.get_entry_object_from_uuid(list_box_row.get_uuid()))
             self.pathbar.add_pathbar_button_to_pathbar(list_box_row.get_uuid())
             self.show_page_of_new_directory(False, False)
@@ -613,60 +550,119 @@ class UnlockedDatabase:
 
     def on_entry_row_button_pressed(self, widget, event):
         self.start_database_lock_timer()
-        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 3 and self.selection_ui.selection_mode_active is False:
-            self.entry_marked_for_delete = self.database_manager.get_entry_object_from_uuid(widget.get_parent().get_uuid())
-            entry_context_popover = self.builder.get_object("entry_context_popover")
-            entry_context_popover.set_relative_to(widget)
-            entry_context_popover.show_all()
-            entry_context_popover.popup()
+        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 3 and self.search.search_active is False:
+            if self.selection_ui.selection_mode_active is False:
+                self.selection_ui.set_selection_headerbar(None, select_row=widget.get_parent())
+            else:
+                self.selection_ui.row_selection_toggled(widget.get_parent())
 
-    def on_entry_delete_menu_button_clicked(self, action, param):
+    def on_element_delete_menu_button_clicked(self, action, param):
         self.start_database_lock_timer()
-        entry_uuid = self.database_manager.get_entry_uuid_from_entry_object(self.entry_marked_for_delete)
+
+        element_to_delete = self.current_group
+        uuid = NotImplemented
+        if self.database_manager.check_is_group_object(self.current_group) is True:
+            uuid = self.database_manager.get_group_uuid_from_group_object(self.current_group)
+            self.current_group = self.database_manager.get_group_parent_group_from_object(self.current_group)
+            self.database_manager.delete_group_from_database(element_to_delete)
+        else:
+            uuid = self.database_manager.get_entry_uuid_from_entry_object(self.current_group)
+            self.current_group = self.database_manager.get_entry_parent_group_from_entry_object(self.current_group)
+            self.database_manager.delete_entry_from_database(element_to_delete)
 
         # If the deleted entry is in the pathbar, we need to rebuild the pathbar
-        if self.pathbar.is_pathbar_button_in_pathbar(entry_uuid) is True:
+        if self.pathbar.is_pathbar_button_in_pathbar(uuid) is True:
             self.pathbar.rebuild_pathbar(self.current_group)
 
-        if self.entry_marked_for_delete is not None:
-            self.database_manager.delete_entry_from_database(self.entry_marked_for_delete)
         self.update_current_stack_page()
+        self.show_page_of_new_directory(False, False)
+
+    def on_entry_duplicate_menu_button_clicked(self, action, param):
+        self.start_database_lock_timer()
+
+        self.database_manager.duplicate_entry(self.current_group)
+        parent_group = self.database_manager.get_entry_parent_group_from_entry_object(self.current_group)
+
+        if self.database_manager.check_is_root_group(parent_group) is True:
+            self.pathbar.on_home_button_clicked(self.pathbar.home_button)
+        else:
+            for button in self.pathbar:
+                if button.get_name() == "PathbarButtonDynamic" and type(button) is passwordsafe.pathbar_button.PathbarButton:
+                    if button.uuid == self.database_manager.get_group_uuid_from_group_object(parent_group):
+                        self.pathbar.on_pathbar_button_clicked(button)
+
+        self.current_group = parent_group
+        self.update_current_stack_page()
+        self.show_page_of_new_directory(False, False)
 
     def on_group_row_button_pressed(self, widget, event):
         self.start_database_lock_timer()
-        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 3 and self.selection_ui.selection_mode_active is False:
-            self.group_marked_for_delete = self.database_manager.get_group_object_from_uuid(widget.get_parent().get_uuid())
-            self.group_marked_for_edit = self.database_manager.get_group_object_from_uuid(widget.get_parent().get_uuid())
-            group_context_popover = self.builder.get_object("group_context_popover")
-            group_context_popover.set_relative_to(widget)
-            group_context_popover.show_all()
-            group_context_popover.popup()
+        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 3 and self.search.search_active is False:
+            if self.selection_ui.selection_mode_active is False:
+                self.selection_ui.set_selection_headerbar(None, select_row=widget.get_parent())
+            else:
+                self.selection_ui.row_selection_toggled(widget.get_parent())
 
-    def on_group_delete_menu_button_clicked(self, action, param):
+    def on_group_edit_button_clicked(self, button):
         self.start_database_lock_timer()
-        group_uuid = self.database_manager.get_group_uuid_from_group_object(self.group_marked_for_delete)
+        group_uuid = button.get_parent().get_parent().get_parent().get_parent().get_uuid()
+        group_object = self.database_manager.get_group_object_from_uuid(group_uuid)
 
-        # If the deleted group is in the pathbar, we need to rebuild the pathbar
-        if self.pathbar.is_pathbar_button_in_pathbar(group_uuid) is True:
-            self.pathbar.rebuild_pathbar(self.current_group)
-
-        self.database_manager.delete_group_from_database(self.group_marked_for_delete)
-        self.update_current_stack_page()
-
-    def on_group_edit_menu_button_clicked(self, action, param):
-        self.start_database_lock_timer()
-        group_uuid = self.database_manager.get_entry_uuid_from_entry_object(self.group_marked_for_edit)
-
-        self.set_current_group(self.group_marked_for_edit)
+        self.set_current_group(group_object)
         self.pathbar.add_pathbar_button_to_pathbar(group_uuid)
         self.show_page_of_new_directory(True, False)
 
     def on_copy_secondary_button_clicked(self, widget, position, eventbutton):
+        self.send_to_clipboard(widget.get_text())
+
+    def send_to_clipboard(self, text):
         self.start_database_lock_timer()
         if self.clipboard_timer is not NotImplemented:
             self.clipboard_timer.cancel()
 
-        self.clipboard.set_text(widget.get_text(), -1)
+        replace_string = text
+        for ref in re.finditer("({REF:.*?})", text):
+            not_valid = False
+            code = ref.group()[5]
+
+            try:
+                uuid = self.hex_to_base64(self.reference_to_hex_uuid(ref.group()))
+            except(Exception):
+                not_valid = True
+
+            value = NotImplemented
+
+            if not_valid is False:
+                if code == "T":
+                    try:
+                        value = self.database_manager.get_entry_name_from_entry_uuid(uuid)
+                    except(AttributeError):
+                        value = ref.group()
+                elif code == "U":
+                    try:
+                        value = self.database_manager.get_entry_username_from_entry_uuid(uuid)
+                    except(AttributeError):
+                        value = ref.group()
+                elif code == "P":
+                    try:
+                        value = self.database_manager.get_entry_password_from_entry_uuid(uuid)
+                    except(AttributeError):
+                        value = ref.group()
+                elif code == "A":
+                    try:
+                        value = str(self.database_manager.get_entry_url_from_entry_uuid(uuid))
+                    except(AttributeError):
+                        value = ref.group()
+                elif code == "N":
+                    try:
+                        value = str(self.database_manager.get_entry_notes_from_entry_uuid(uuid))
+                    except(AttributeError):
+                        value = ref.group()
+
+                replace_string = replace_string.replace(ref.group(), value)
+
+        self.clipboard.set_text(replace_string, -1)
+
         self.show_database_action_revealer(_("Copied to clipboard"))
         clear_clipboard_time = passwordsafe.config_manager.get_clear_clipboard()
         self.clipboard_timer = Timer(clear_clipboard_time, self.clear_clipboard)
@@ -737,13 +733,17 @@ class UnlockedDatabase:
 
         save_dialog.present()
 
+    def show_references_dialog(self, action, param):
+        ReferencesDialog(self)
+
+    def show_properties_dialog(self, action, param):
+        PropertiesDialog(self)
+
     #
     # Utils
     #
 
     def show_database_action_revealer(self, message):
-        database_action_box = self.builder.get_object("database_action_box")
-
         database_action_label = self.builder.get_object("database_action_label")
         database_action_label.set_text(message)
 
@@ -765,6 +765,15 @@ class UnlockedDatabase:
         if self.database_settings_dialog is not NotImplemented:
             self.database_settings_dialog.close()
 
+        if self.notes_dialog is not NotImplemented:
+            self.notes_dialog.close()
+
+        if self.references_dialog is not NotImplemented:
+            self.references_dialog.close()
+
+        if self.properties_dialog is not NotImplemented:
+            self.properties.close()
+
         if passwordsafe.config_manager.get_save_automatically() is True:
             save_thread = threading.Thread(target=self.database_manager.save_database)
             save_thread.daemon = False
@@ -785,6 +794,12 @@ class UnlockedDatabase:
         if self.database_settings_dialog is not NotImplemented:
             self.database_settings_dialog.close()
 
+        if self.references_dialog is not NotImplemented:
+            self.references_dialog.close()
+
+        if self.properties_dialog is not NotImplemented:
+            self.properties.close()
+
         if passwordsafe.config_manager.get_save_automatically() is True:
             save_thread = threading.Thread(target=self.database_manager.save_database)
             save_thread.daemon = False
@@ -870,6 +885,15 @@ class UnlockedDatabase:
     def remove_loading_indicator(self, overlay, spinner):
         overlay.remove(spinner)
 
+    def reference_to_hex_uuid(self, reference_string):
+        return reference_string[9:-1].lower()
+
+    def hex_to_base64(self, hex_uuid):
+        return codecs.encode(codecs.decode(hex_uuid.encode(), 'hex'), 'base64').decode().splitlines()[0]
+
+    def base64_to_hex(self, base64_uuid):
+        return codecs.encode(codecs.decode(base64_uuid.encode(), 'base64'), 'hex').decode().upper()
+
     #
     # DBus
     #
